/*
 * Copyright (c) 2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Tue Feb 28 2017
 */
#include "acarduio.h"

#include <err.h>
#include <fnmatch.h>
#include <stdio.h>
#include <unistd.h>

#include "arduio_c_types.h"
#include "codels.h"


/* --- Task comm -------------------------------------------------------- */


/** Codel ario_comm_start of task comm.
 *
 * Triggered by arduio_start.
 * Yields to arduio_poll.
 * Throws arduio_e_sys.
 */
genom_event
ario_comm_start(arduio_conn_s **conn, const arduio_ios *ios,
                const genom_context self)
{
  arduio_state_s *data;

  *conn = malloc(sizeof(**conn));
  if (!*conn) return ario_e_sys_error(NULL, self);

  (*conn)->chan.fd = -1;
  (*conn)->chan.r = (*conn)->chan.w = 0;
  (*conn)->chan.start = false;

  data = ios->data(self);
  data->diodir._length = 0;
  data->digital._length = 0;
  data->analog._length = 0;
  ios->write(self);

  return arduio_poll;
}


/** Codel ario_comm_poll of task comm.
 *
 * Triggered by arduio_poll.
 * Yields to arduio_poll, arduio_err, arduio_recv.
 * Throws arduio_e_sys.
 */
genom_event
ario_comm_poll(const arduio_conn_s *conn, const genom_context self)
{
  int s;

  s = ario_wait_msg(&conn->chan);
  if (s < 0) {
    if (errno != EINTR) return arduio_err;
    return arduio_poll;
  }
  else if (s == 0) return arduio_poll;

  return arduio_recv;
}


/** Codel ario_comm_recv of task comm.
 *
 * Triggered by arduio_recv.
 * Yields to arduio_poll, arduio_recv.
 * Throws arduio_e_sys.
 */
genom_event
ario_comm_recv(arduio_conn_s **conn, double ioscale[16],
               const arduio_ios *ios, const genom_context self)
{
  arduio_state_s *data;
  uint8_t *msg, len;
  uint16_t v;
  uint8_t i;

  if (ario_recv_msg(&(*conn)->chan, false) != 1) {
    ios->write(self);
    return arduio_poll;
  }

  data = ios->data(self);
  msg = (*conn)->chan.msg;
  len = (*conn)->chan.len;
  switch(*msg++) {
    case 'D': /* GPIO dir state array */
      if (len > data->diodir._maximum) len = data->diodir._maximum;
      for(i = 0; i < len-1; i++) {
        switch(*msg++) {
          case 0: data->diodir._buffer[i] = arduio_IO_IN; break;
          case 1: data->diodir._buffer[i] = arduio_IO_IN_PU; break;
          case 2: data->diodir._buffer[i] = arduio_IO_OUT; break;
          case 3: data->diodir._buffer[i] = arduio_IO_PWM_31kHz; break;
          case 4: data->diodir._buffer[i] = arduio_IO_PWM_122Hz; break;
          case 5: data->diodir._buffer[i] = arduio_IO_PWM_31Hz; break;
          case 16: data->diodir._buffer[i] = arduio_IO_WS28XX; break;
        }

        if (data->diodir._buffer[i] == arduio_IO_WS28XX)
          ioscale[i] = 1.;
        else
          ioscale[i] = 65535.;
      }
      data->diodir._length = i;
      break;

    case 'I': /* GPIO state array */
      len = (len - 1)/2;
      if (len >= data->digital._maximum) len = data->digital._maximum - 1;
      for(i = 0; i < len; i++) {
        v = *msg++ << 8;
        v += *msg++;
        data->digital._buffer[i] = v / ioscale[i];
      }
      data->digital._length = i;
      break;

    case 'i': /* GPIO state */
      if (len == 4) {
        i = *msg++;
        if (i < data->digital._maximum) {
          v = *msg++ << 8;
          v += *msg++;
          data->digital._buffer[i] = v / ioscale[i];
          if (data->digital._length < i+1)
            data->digital._length = i+1;
        }
      }
      break;

    case 'A': /* analog state array */
      len = (len - 1)/2;
      if (len >= data->analog._maximum) len = data->analog._maximum - 1;
      for(i = 0; i < len; i++) {
        v = *msg++ << 8;
        v += *msg++;
        data->analog._buffer[i] = v / 65535.;
      }
      data->analog._length = i;
      break;

    case 'a': /* analog state */
      if (len == 4) {
        i = *msg++;
        if (i < data->analog._maximum) {
          v = *msg++ << 8;
          v += *msg++;
          data->analog._buffer[i] = v / 65535.;
          if (data->analog._length < i+1)
            data->analog._length = i+1;
        }
      }
      break;
  }

  return arduio_recv;
}


/** Codel ario_comm_err of task comm.
 *
 * Triggered by arduio_err.
 * Yields to arduio_poll.
 * Throws arduio_e_sys.
 */
genom_event
ario_comm_err(arduio_conn_s **conn, const genom_context self)
{
  if ((*conn)->chan.fd >= 0) {
    close((*conn)->chan.fd);
    (*conn)->chan.fd = -1;
    warnx("disconnected from %s", (*conn)->chan.path);
  }
  return arduio_poll;
}


/** Codel ario_comm_stop of task comm.
 *
 * Triggered by arduio_stop.
 * Yields to arduio_ether.
 * Throws arduio_e_sys.
 */
genom_event
ario_comm_stop(arduio_conn_s **conn, const genom_context self)
{
  if (!*conn) return arduio_ether;

  if ((*conn)->chan.fd >= 0) close((*conn)->chan.fd >= 0);
  free(*conn);
  *conn = NULL;
  return arduio_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel ario_connect_start of activity connect.
 *
 * Triggered by arduio_start.
 * Yields to arduio_ether.
 * Throws arduio_e_sys, arduio_e_baddev.
 */
genom_event
ario_connect_start(const char serial[64], arduio_conn_s **conn,
                   const genom_context self)
{
  static const char magic[] = "[?]arduio1.1*";

  int c, s;

  if ((*conn)->chan.fd >= 0) {
    close((*conn)->chan.fd);
    warnx("disconnected from %s", (*conn)->chan.path);
  }

  /* open tty */
  (*conn)->chan.fd = ario_open_tty(serial);
  if ((*conn)->chan.fd < 0) return ario_e_sys_error(serial, self);

  /* check endpoint */
  while (ario_recv_msg(&(*conn)->chan, true) == 1); /* flush buffer */
  do {
    c = 0;
    do {
      if (ario_send_msg(&(*conn)->chan, "?")) /* ask for id */
        return ario_e_sys_error(serial, self);

      s = ario_wait_msg(&(*conn)->chan);
      if (s < 0 && errno != EINTR) return ario_e_sys_error(NULL, self);
      if (s > 0) break;
    } while(c++ < 6);
    if (c > 6) {
      errno = ETIMEDOUT;
      return ario_e_sys_error(NULL, self);
    }

    s = ario_recv_msg(&(*conn)->chan, true);
  } while(s == 1 && (*conn)->chan.msg[0] != '?');

  (*conn)->chan.msg[(*conn)->chan.len] = 0;
  if (s != 1 || fnmatch(magic, (char *)(*conn)->chan.msg, 0)) {
    if (s != 1) {
      errno = ENOMSG;
      return ario_e_sys_error(NULL, self);
    } else {
      arduio_e_baddev_detail d;
      snprintf(d.dev, sizeof(d.dev), "device is `%s' instead of `%s'",
               (*conn)->chan.msg, magic);
      close((*conn)->chan.fd);
      (*conn)->chan.fd = -1;
      return arduio_e_baddev(&d, self);
    }
  }

  snprintf((*conn)->chan.path, sizeof((*conn)->chan.path), "%s", serial);
  warnx("connected to %s", (*conn)->chan.path);

  ario_send_msg(&(*conn)->chan, "d"); /* get current dir state */
  return arduio_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel ario_disconnect_start of activity disconnect.
 *
 * Triggered by arduio_start.
 * Yields to arduio_ether.
 * Throws arduio_e_sys.
 */
genom_event
ario_disconnect_start(arduio_conn_s **conn, const genom_context self)
{
  ario_comm_err(conn, self);
  return arduio_ether;
}
